Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq path div choice.
Require Import fintype finset bigop binomial prime ssrint ssralg ssrnum.

Require Import rat.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Import GRing.Theory.
Import Num.Theory.

(******************************************************************************)
(* The main purpose of this library is to establish properties of the         *)
(* sequence lcm(1..n). The proof of irrationality that we formalize           *)
(* relies on lemma 'dvdnbinD' proved in the present file, which               *)
(* states a strong divisibility property between binomial coefficients and    *)
(* lcm(1..n).                                                                 *)
(* This file defines:                                                         *)
(* iter_lcmn n   == the least common multiple of the n-th first natural       *)
(*                  numbers                                                   *)
(* log_floor p n == the largest power of p that is smaller or equal to n      *)
(******************************************************************************)


(**** Some extra results about (the factorial of) the p-valuation of a ****)
(**** natural number, implemented by (logn p n) with p prime.          ****)

Lemma fact_logp_sum p n : prime p -> logn p n`! = \sum_(i < n) (n %/ p ^ i.+1).
Proof.
elim: n => [|n ihn] p_prime; first by rewrite fact0 logn1 big_ord0.
rewrite factS lognM ?fact_gt0 // big_ord_recr /= ihn //=.
rewrite divn_small ?addn0; last by rewrite ltn_expl ?prime_gt1.
have -> : logn p n.+1 = \sum_(i < n) (p ^ i.+1 %| n.+1).
  rewrite -big_mkcond sum1dep_card; set P := [set x | _].
  have -> : P = [set i : 'I__ | i.+1 <= logn p n.+1].
    apply/setP => i; rewrite !inE /= -{1}(@partnC p n.+1) //.
    rewrite Gauss_dvdl; first by rewrite p_part dvdn_Pexp2l ?prime_gt1.
    by rewrite (@pnat_coprime p) ?part_pnat ?pnat_exp ?pnat_id.
  rewrite -sum1dep_card big_ord_narrow ?sum_nat_const ?muln1 ?card_ord //.
  rewrite -(@leq_exp2l p) ?prime_gt1 // -p_part.
  by rewrite (leq_trans (dvdn_leq _ (dvdn_part p _))) // ltn_expl ?prime_gt1.
rewrite -big_split; apply: eq_big => //= i _; set q := p ^ i.+1; symmetry.
have q_gt0 : 0 < q by rewrite ?expn_gt0 ?prime_gt0.
rewrite {1}(divn_eq n q) -addn1 -addnA divnMDl // addnC addn1.
have -> // : (q %| n.+1) = ((n %% q).+1 == q).
  rewrite -{3}[q]prednK // eqSS.
  by rewrite -(@modn_small q.-1 q) -1?(@eqn_modDr 1) ?addn1 prednK // modnn.
have [->|hi] := altP eqP; first by rewrite divnn q_gt0.
by rewrite divn_small // ltn_neqAle hi ltn_mod.
Qed.

(* Same as fact_logp_sum, with a wider range for the sum. *)
Lemma fact_logp_sum_widen p m n : prime p -> (m >= n)%N ->
  logn p n`! =  \sum_(i < m) (n %/ p ^ i.+1).
move => prime_p leq_nm; rewrite logn_fact //; symmetry.
(* FIXME : here I would like to simply 'rewrite big_mkord', but I need *)
(* to specify both the (dummy) predicate and the summand... *)
pose G (i : nat) := n %/ p ^ i.+1.
rewrite -[LHS](big_mkord (fun _ => true) G); rewrite {}/G.
rewrite (big_cat_nat _ _ _ (leq0n n) leq_nm) /= addnC.
rewrite big_nat_cond /= big1; first by rewrite add0n big_add1.
move=> i; rewrite andbT => /andP [leni ltim]; apply: divn_small.
apply: leq_trans (ltn_expl _ _); first exact: leq_trans leni _.
exact: prime_gt1.
Qed.

(* Yet an other variant of fact_logp_sum. *)
Lemma fact_logp_sum_small p j n : prime p -> (n < p ^ j.+1) ->
  logn p n`! =  \sum_(i < j) (n %/ p ^ i.+1).
Proof.
move => prime_p leq_npj.
case: (ltnP j n) => [le_jn|lt_nj]; last by rewrite (@fact_logp_sum_widen p j).
rewrite logn_fact //. have hl : 1 <= j.+1 by [].
have hr : j.+1 <= n.+1 by exact: ltnW.
rewrite (big_cat_nat _ _ _ hl hr) /= addnC big_nat_cond /= big1.
  by rewrite add0n big_add1 /= big_mkord.
move=> i; rewrite andbT => /andP [leni ltim]; apply: divn_small.
apply: leq_trans leq_npj _; rewrite leq_exp2l //; exact: prime_gt1.
Qed.

(**** Definition and some elementary results about lcm(1..n) ****)


(* This is the least common multiple of the n-th first natural numbers *)
Definition iter_lcmn (n : nat) := \big[lcmn/1%N]_(1 <= i < n.+1) i.

(* We use the same notation as in the paper. *)
Local Notation l := iter_lcmn.

Fact iter_lcmn_gt0 (n : nat) : (n >= 1)%N -> (l n > 0)%N.
Proof.
case: n => // n _; elim: n => [|n ihn]; first by rewrite /iter_lcmn big_nat1.
by rewrite /iter_lcmn (@big_cat_nat _ _ _ n.+2) //= lcmn_gt0 ihn big_nat1.
Qed.

Fact iter_lcmn_leq_div (n m : nat) : (n <= m)%N -> l n %| l m.
Proof.
move=> ?; rewrite /iter_lcmn [X in _ %| X](@big_cat_nat _ _ _ n.+1) //. 
exact: dvdn_lcml.
Qed.

Fact iter_lcmn_div (n m : nat) : (0 < n)%N -> (n <= m)%N -> n %| l m.
Proof.
move=> lt0n lenm; apply: dvdn_trans (iter_lcmn_leq_div lenm); rewrite /iter_lcmn.
rewrite (@big_cat_nat _ _ _ n) //= big_nat1; exact: dvdn_lcmr.
Qed.

(**** Definition and some elementary results about log_floor. ****)

(* We need the two following facts in order to define log_floor as a maximum. *)
Fact ex_log_floor (p k : nat) : exists j, p ^ j <= k.+1.
Proof. by exists 0; rewrite expn0. Qed.

Fact maj_log_floor (p k j : nat) : p.+2 ^ j <= k.+1 -> j <= k.+1.
Proof.
move=> hpj; apply: leq_trans hpj; exact: leq_trans (ltn_expl _  _).
Qed.

(* The actual definition of log_floor. *)
Definition log_floor (p k : nat) : nat :=
  if p is p'.+2 then
    if k is k'.+1 then ex_maxn (@ex_log_floor p'.+2 k') (@maj_log_floor p' k')
      else 0
  else k.

(* The maximallity properties of log_floor *)
Fact log_floorP (p n : nat) : p >= 2 -> n >= 1 -> p ^ (log_floor p n) <= n.
Proof.
case: p => // [] [|p] // _; case: n => [|n] // _.
by rewrite /log_floor; case: ex_maxnP.
Qed.

Fact log_floor_max (p k j : nat) : p >= 2 -> p ^ j <= k -> j <= log_floor p k.
Proof.
case: p => //; case => // p _ ; case: k; first by rewrite leqn0 expn_eq0.
by move=> k hp1; rewrite /log_floor; case: ex_maxnP=> i hmax; apply.
Qed.

Fact log_floor_ltn (p n : nat) : p >= 2 -> n < p ^ (log_floor p n).+1.
Proof.
move=> hp; apply: contraLR hp; rewrite -ltnNge; case: (leqP 2 p) => hp2 //.
by rewrite ltnS; move/(log_floor_max hp2); rewrite ltnn.
Qed.

(* (log_floor p) is greater than (logn p) in the following sense: *)
Lemma  leq_logn_floor p m k : (1 < p)%N -> (m <= k)%N -> 
                              (logn p m <= log_floor p k)%N.
Proof.
case: m => [|m]; rewrite ?logn0 //.
move=> h1p leq_mk; apply: log_floor_max=> //; apply: leq_trans leq_mk.
by rewrite -p_part dvdn_leq // dvdn_part.
Qed.

(* (logn p n!) can be expressed as a sum of log_floor *)
Lemma logp_sum_floor p n : prime p ->
 logn p (n`!) =  \sum_(i < log_floor p n) (n %/ p ^ i.+1).
Proof.
move=> hp; rewrite -fact_logp_sum_small //; apply: log_floor_ltn. 
exact:  prime_gt1.
Qed.

(**** p-Valuations of binomial coefficients ****)

(* This result about p-valuations of binomial coefficients and lcn(1..n) *)
(* was the motivation for introducing log_floor. *)
(* FIXME : the proof probably needs to be cleaned up. *)
Lemma bin_valp m k p : (m > 0) -> prime p -> (m <= k) ->
                       (logn p (m * 'C(k, m)) <= logn p (l k)).
Proof.
move => gt_m_0 prime_p leq_mk.
have hp2 : (p >= 2) by exact: prime_gt1.
rewrite lognM ?bin_gt0 //.
have gt_k0 : (k > 0) by rewrite (@leq_trans m).
pose j := log_floor p k; apply: (@leq_trans j); rewrite /j; last first.
  rewrite - pfactor_dvdn ?d_lt0 // {j}; last by apply: iter_lcmn_gt0.
  apply: iter_lcmn_div; [by rewrite expn_gt0 ltnW | exact: log_floorP].
rewrite -lez_nat PoszD bin_factd //. (* going to int eases later \sum handling *)
have logn_divz a b : (a > 0)%N -> b %| a ->
  ((logn p (a %/ b))%:Z = (logn p a)%:Z - (logn p b)%:Z)%R.
   by move=> ? ?; rewrite logn_div // -subzn // dvdn_leq_log.
rewrite {}logn_divz ?fact_gt0 //; first last.
   by apply/dvdnP; exists 'C(k,m); rewrite bin_fact.
rewrite lognM ?fact_gt0 // logp_sum_floor //.
rewrite  !(@fact_logp_sum_small p (log_floor p k)) //; first last.
- by apply: (leq_ltn_trans leq_mk); apply: log_floor_ltn.
- by apply  (leq_ltn_trans (leq_subr m k) (log_floor_ltn k hp2)).
rewrite -!big_split /= !(big_morph _ PoszD (erefl _)) /= -sumrB -ler_sub_addl.
have hsplit : log_floor p k = (logn p m) + ((log_floor p k) - (logn p m)).
  by rewrite subnKC //; exact: (leq_logn_floor _ _).
rewrite {}[in X in (X <= _)%R]hsplit big_split_ord /= big1 ?add0r /=; last first.
  have hfirst i : (p ^ i %| m) -> k %/ p ^ i = m %/ p ^ i + (k - m) %/ p ^ i.
    by move=> h; rewrite -{1}[k](@subnK m) // divnDr // addnC.
  by case=> i hi _ /=; rewrite hfirst ?subrr // pfactor_dvdn.
set rhs := (X in (_ <= X)%R).
have {rhs} -> /= : rhs = \sum_(i < (log_floor p k) - logn p m) 1 :> int.
  by rewrite /rhs sum1_card card_ord subzn // leq_logn_floor.
rewrite !(big_morph _ PoszD (erefl _) ) /=; apply: ler_sum.
move => i _; rewrite PoszD opprD addrA -ler_sub_addr opprK subzn; last first.
  by rewrite -lez_nat; apply: leq_div2r.
by rewrite -PoszD lez_nat [1 + _]addnC; apply: leq_trans (geq_divBl _ _ _) _.
Qed.

(* In the proof that zeta(3) is irrational, we use this corollary to show *)
(* that (lcn (1..n))^3 * b_n is an integer. *)
Corollary dvd_nbin_iter_lcmn (i j : nat) (n : nat) : 
  (1 <= j) -> (j  <= i) -> (i <= n) -> (j * 'C(i, j) %| l n).
Proof.
move=> le0j leji lein. 
apply/dvdn_partP => [|p]; first by rewrite muln_gt0 bin_gt0 le0j.
move=> hp; apply: dvdn_trans ( iter_lcmn_leq_div lein); move: hp.
rewrite mem_primes; case/and3P=> pp hpos hdvd.
rewrite p_part pfactor_dvdn //; last first. 
  by apply: iter_lcmn_gt0; apply: leq_trans leji.
exact: bin_valp.
Qed.
